from importlib.metadata import version
from typing import Tuple, Type

from typing_extensions import Self
from web3.eth import BaseEth
from web3._utils.blocks import select_method_for_block_identifier
from web3._utils.rpc_abi import RPC
from web3.method import Method, TFunc, default_root_munger
from web3.types import BlockIdentifier, RPCEndpoint

from dank_mids._web3.formatters import (
    _response_formatters,
    _get_response_formatters,
    get_request_formatters,
)

WEB3_MAJOR_VERSION = int(version("web3").split(".")[0])


class MethodNoFormat(Method[TFunc]):
    """Custom method class to bypass web3py's default result formatters.

    This class processes parameters, makes conditional adjustments to the parameters,
    and selects specific methods to call based on those parameters. It bypasses the
    standard result formatters by applying a no-operation formatter, effectively
    returning responses as-is or using predefined formatters based on the RPC call.
    """

    def process_params(self, module, *args, **kwargs):
        """Prepare and return request parameters and response formatters.

        Adjusts parameters based on the method choice logic and avoids standard request
        param forwarding for specific filter methods. Defines response formatters to
        bypass usual formatting processes.

        Args:
            module: The module where the method belongs.
            args: Positional arguments for the method.
            kwargs: Keyword arguments for the method.
        """
        params = self.input_munger(module, args, kwargs)

        if self.method_choice_depends_on_args:
            self.json_rpc_method = self.method_choice_depends_on_args(value=params[0])

            pending_or_latest_filter_methods = (
                RPC.eth_newPendingTransactionFilter,
                RPC.eth_newBlockFilter,
            )
            if self.json_rpc_method in pending_or_latest_filter_methods:
                params = []

        method = self.method_selector_fn()
        request_formatters = get_request_formatters(method)
        if method == "eth_call":
            # we will only format the TxParams
            if len(params) > 2:
                raise NotImplementedError("You can not use state override with dank mids")
            tx, block = params
            if len(tx) > 2:
                params = request_formatters(params)
            else:
                params = tx, hex(block) if isinstance(block, int) else block
                if isinstance(tx["data"], bytes):
                    tx["data"] = f"0x{bytes.hex(tx['data'])}"
        else:
            params = request_formatters(params)
        request = method, params
        return request, _response_formatters.get(self.json_rpc_method) or _get_response_formatters(
            self.json_rpc_method
        )

    @classmethod
    def default(cls, method: RPCEndpoint) -> Self:
        """Creates a default instance of MethodNoFormat.

        Args:
            method: The RPC method for which a MethodNoFormat instance should be created.
        """
        return cls(method, [default_root_munger])


def bypass_chainid_formatter(eth: Type[BaseEth]) -> None:
    """Bypasses the formatter for the eth_chainId method.

    Args:
        eth: The Ethereum base class instance whose method is to be modified.
    """
    eth._chain_id = MethodNoFormat(RPC.eth_chainId)


def bypass_getbalance_formatter(eth: Type[BaseEth]) -> None:
    """Bypasses the formatter for the eth_getBalance method.

    Args:
        eth: The Ethereum base class instance whose method is to be modified.
    """
    eth._get_balance = MethodNoFormat(RPC.eth_getBalance, mungers=[eth.block_id_munger])


def bypass_transaction_count_formatter(eth: Type[BaseEth]) -> None:
    """Bypasses the formatter for the eth_getTransactionCount method.

    Args:
        eth: The Ethereum base class instance whose method is to be modified.
    """
    eth._get_transaction_count = MethodNoFormat(
        RPC.eth_getTransactionCount, mungers=[eth.block_id_munger]
    )


def bypass_log_formatter(eth: Type[BaseEth]) -> None:
    """Bypasses formatters for log-related methods such as eth_getLogs.

    Args:
        eth: The Ethereum base class instance whose methods are to be modified.
    """
    eth._get_logs = MethodNoFormat.default(RPC.eth_getLogs)
    eth._get_logs_raw = MethodNoFormat.default(f"{RPC.eth_getLogs}_raw")
    eth.get_filter_logs = MethodNoFormat.default(RPC.eth_getFilterLogs)
    eth.get_filter_changes = MethodNoFormat.default(RPC.eth_getFilterChanges)


def bypass_transaction_receipt_formatter(eth: Type[BaseEth]) -> None:
    """Bypasses the formatter for the eth_getTransactionReceipt method.

    Args:
        eth: The Ethereum base class instance whose method is to be modified.
    """
    method = MethodNoFormat.default(RPC.eth_getTransactionReceipt)
    if WEB3_MAJOR_VERSION >= 6:
        eth._transaction_receipt = method
    else:
        eth._get_transaction_receipt = method


def bypass_transaction_formatter(eth: Type[BaseEth]) -> None:
    """Bypasses the formatter for the eth_getTransactionByHash method.

    Args:
        eth: The Ethereum base class instance whose method is to be modified.
    """
    eth._get_transaction = MethodNoFormat.default(RPC.eth_getTransactionByHash)


_block_selectors = dict(
    if_predefined=RPC.eth_getBlockByNumber,
    if_hash=RPC.eth_getBlockByHash,
    if_number=RPC.eth_getBlockByNumber,
)


def bypass_block_formatters(eth: Type[BaseEth]) -> None:
    """Bypasses the formatter for block-related methods such as eth_getBlockByNumber.

    Adjusts the mungers definition based on the major version of the web3 library.

    Args:
        eth: The Ethereum base class instance whose methods are to be modified.
    """
    if WEB3_MAJOR_VERSION >= 6:
        get_block_munger = eth.get_block_munger
    else:

        def get_block_munger(
            self, block_identifier: BlockIdentifier, full_transactions: bool = False
        ) -> Tuple[BlockIdentifier, bool]:
            return (block_identifier, full_transactions)

    eth._get_block = MethodNoFormat(
        method_choice_depends_on_args=select_method_for_block_identifier(**_block_selectors),
        mungers=[get_block_munger],
    )


def bypass_eth_call_formatter(eth: Type[BaseEth]) -> None:
    """Bypasses the formatter for the eth_call method.

    Args:
        eth: The Ethereum base class instance whose method is to be modified.
    """
    eth._call = MethodNoFormat(RPC.eth_call, mungers=[eth.call_munger])


def bypass_get_code_formatter(eth: Type[BaseEth]) -> None:
    """Bypasses the formatter for the eth_getCode method.

    Args:
        eth: The Ethereum base class instance whose method is to be modified.
    """
    eth._get_code = MethodNoFormat(RPC.eth_getCode, mungers=[eth.block_id_munger])


skip_formatters = (
    bypass_chainid_formatter,
    bypass_getbalance_formatter,
    bypass_transaction_count_formatter,
    bypass_eth_call_formatter,
    bypass_get_code_formatter,
    bypass_log_formatter,
    bypass_transaction_receipt_formatter,
    bypass_transaction_formatter,
    bypass_block_formatters,
)


def bypass_formatters(eth) -> None:
    """Executes a sequence of bypass methods to remove default formatters for ETH methods.

    Utilizes a predefined list of bypass methods to systematically modify the corresponding
    methods in the Ethereum base class instance.

    Args:
        eth: The Ethereum base class instance whose methods are to be modified.
    """
    for bypass in skip_formatters:
        bypass(eth)



"""eth_utils uses a decorator to ensure certain functions are called with proper args.

This is helpful for development but adds unnecessary overhead for production use.

This module lets us redefine those functions without the decorator so our scripts can go fast. Vroom.
"""

import binascii
from typing import Any, AnyStr, Final, Optional

from eth_typing import HexStr


hexlify: Final = binascii.hexlify
unhexlify: Final = binascii.unhexlify


def patch_eth_utils() -> None:
    """eth_utils uses a decorator to ensure certain functions are called with proper args.

    This is helpful for development but adds unnecessary overhead for production use.

    This function redefines those functions without the decorator so our scripts can go fast. Vroom.
    """
    import eth_utils.crypto

    eth_utils.crypto.to_bytes = to_bytes

    import web3.main

    web3.main.to_hex = to_hex
    web3.main.to_bytes = to_bytes

    import web3.middleware.filter

    web3.middleware.filter.to_hex = to_hex

    import web3.providers.base

    web3.providers.base.to_bytes = to_bytes

    import web3._utils.encoding

    web3._utils.encoding.to_hex = to_hex
    web3._utils.encoding.to_bytes = to_bytes

    import web3._utils.events

    web3._utils.events.to_hex = to_hex
    web3._utils.events.to_bytes = to_bytes

    import web3._utils.normalizers

    web3._utils.normalizers.to_hex = to_hex
    web3._utils.normalizers.to_bytes = to_bytes

    try:
        # only present on some versions
        import web3._utils.type_conversion
    except ModuleNotFoundError:
        pass
    else:
        web3._utils.type_conversion.to_hex = to_hex
        web3._utils.type_conversion.to_bytes = to_bytes


def to_hex(
    primitive: Optional[Any] = None,
    hexstr: Optional[HexStr] = None,
    text: Optional[str] = None,
) -> HexStr:
    """
    Auto converts any supported value into its hex representation.
    Trims leading zeros, as defined in:
    https://github.com/ethereum/wiki/wiki/JSON-RPC#hex-value-encoding
    """
    if hexstr is not None:
        return add_0x_prefix(hexstr.lower())  # type: ignore [arg-type]

    if text is not None:
        return encode_hex(text.encode("utf-8"))

    elif isinstance(primitive, bool):
        return "0x1" if primitive else "0x0"  # type: ignore [return-value]

    elif isinstance(primitive, (bytes, bytearray)):
        return encode_hex(primitive)  # type: ignore [type-var]

    elif isinstance(primitive, memoryview):
        return encode_hex(bytes(primitive))

    elif isinstance(primitive, str):
        raise TypeError(
            "Unsupported type: The primitive argument must be one of: bytes,"
            "bytearray, int or bool and not str"
        )

    elif isinstance(primitive, int):
        return hex(primitive)  # type: ignore [return-value]

    raise TypeError(
        f"Unsupported type: '{repr(type(primitive))}'. Must be one of: bool, str, "
        "bytes, bytearray or int."
    )


def to_bytes(
    primitive: Optional[Any] = None,
    hexstr: Optional[HexStr] = None,
    text: Optional[str] = None,
) -> bytes:
    if isinstance(primitive, bool):
        return b"\x01" if primitive else b"\x00"
    elif isinstance(primitive, (bytearray, memoryview)):
        return bytes(primitive)
    elif isinstance(primitive, bytes):
        return primitive
    elif isinstance(primitive, int):
        return to_bytes(hexstr=to_hex(primitive))
    elif hexstr is not None:
        return decode_hex(f"0x0{remove_0x_prefix(hexstr)}" if len(hexstr) % 2 else hexstr)
    elif text is not None:
        return text.encode("utf-8")
    raise TypeError(
        "expected a bool, int, byte or bytearray in first arg, " "or keyword of hexstr or text"
    )


def decode_hex(value: str) -> bytes:
    if not isinstance(value, str):
        raise TypeError("Value must be an instance of str")
    non_prefixed = remove_0x_prefix(value)  # type: ignore [arg-type]
    # unhexlify will only accept bytes type someday
    ascii_hex = non_prefixed.encode("ascii")
    return unhexlify(ascii_hex)


def encode_hex(value: AnyStr) -> HexStr:
    ascii_bytes: Any  # no need to validate bytes type in the c code
    if isinstance(value, bytes):
        ascii_bytes = value
    elif isinstance(value, str):
        ascii_bytes = value.encode("ascii")
    elif isinstance(value, bytearray):
        ascii_bytes = bytes(value)
    else:
        raise TypeError("Value must be an instance of str or unicode")
    ascii_hex = hexlify(ascii_bytes).decode("ascii")
    return add_0x_prefix(ascii_hex)  # type: ignore [arg-type]


def is_0x_prefixed(value: str) -> bool:
    if not isinstance(value, str):
        raise TypeError(f"is_0x_prefixed requires text typed arguments. Got: {repr(value)}")
    return value.startswith(("0x", "0X"))


def remove_0x_prefix(value: HexStr) -> HexStr:
    # sourcery skip: assign-if-exp, reintroduce-else
    if is_0x_prefixed(value):
        return value[2:]  # type: ignore [return-value]
    return value


def add_0x_prefix(value: HexStr) -> HexStr:
    # sourcery skip: assign-if-exp, reintroduce-else
    if is_0x_prefixed(value):
        return value
    return f"0x{value}"  # type: ignore [return-value]
